// RenderEngine class implementation
#include "pch.h"

#include "RenderEngine.h"
#include "util.h"
#include "ShaderModule.h"

namespace nf::client::render {
    RenderEngine::RenderEngine(std::shared_ptr<Window> window, DisplayConfig display)
        : m_window(window)
        , m_display(display)
        , m_instance()
        , m_surface()
        , m_physicalDevice()
        , m_queueFIGraphics()
        , m_queueFIPresent()
        , m_device()
        , m_queueGraphics()
        , m_queuePresent()
        , m_renderPassOutput()
        , m_swapchain()
        , m_swapchainImageFormat()
        , m_swapchainImages()
        , m_swapchainImageViews()
        , m_swapchainFramebuffers()
        , m_pipelineLayoutOutput()
        , m_pipelineOutput()
        , m_commandPool()
        , m_commandBuffer()
        , m_semaphoreImageAvailable()
        , m_semaphoreRenderFinished()
        , m_fenceInFlight()
    {
        NFLog("Initializing render engine");
        m_window->setDisplay(m_display);

        createInstance();
        createSurface();
        pickPhysicalDevice();
        createDevice();
        createSwapchain();
        createOutputRenderPass();
        createOutputPipeline();
        createCommandBuffer();
        createSynchronizationObjects();

        m_window->show();
    }

    void RenderEngine::createInstance() {
        VkApplicationInfo appInfo = { VK_STRUCTURE_TYPE_APPLICATION_INFO };
        appInfo.apiVersion = VK_API_VERSION_1_0;

        VkInstanceCreateInfo instanceCI = { VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO };
        instanceCI.pApplicationInfo = &appInfo;
#ifdef _DEBUG
        const char* validationLayerName = "VK_LAYER_KHRONOS_validation";
        instanceCI.ppEnabledLayerNames = &validationLayerName;
        instanceCI.enabledLayerCount = 1;
#endif
        std::vector<const char*> instanceExtNames = { VK_KHR_SURFACE_EXTENSION_NAME, VK_KHR_WIN32_SURFACE_EXTENSION_NAME };
        instanceCI.ppEnabledExtensionNames = instanceExtNames.data();
        instanceCI.enabledExtensionCount = instanceExtNames.size();

        if (vkCreateInstance(&instanceCI, nullptr, &m_instance) != VK_SUCCESS)
            NFError("Could not create Vulkan instance.");
    }

    void RenderEngine::createSurface() {
        VkWin32SurfaceCreateInfoKHR surfaceCI = { VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR };
        surfaceCI.hwnd = m_window->getHandle();

        if (vkCreateWin32SurfaceKHR(m_instance, &surfaceCI, nullptr, &m_surface) != VK_SUCCESS)
            NFError("Could not create Vulkan surface.");
    }

    void RenderEngine::pickPhysicalDevice() {
        // TODO: Save GPU choice in DisplayConfig
        // First, get list of all GPUs
        uint32_t numPhysicalDevices = 0;
        vkEnumeratePhysicalDevices(m_instance, &numPhysicalDevices, nullptr);
        if (numPhysicalDevices == 0)
            NFError("No Vulkan GPUs found.");

        std::vector<VkPhysicalDevice> physicalDevices(numPhysicalDevices);
        vkEnumeratePhysicalDevices(m_instance, &numPhysicalDevices, physicalDevices.data());

        // Then gather information on them and save first dedicated
        std::optional<int> firstDedicatedPhysicalDeviceIndex;
        std::vector<VkPhysicalDeviceProperties> physicalDeviceProperties;
        for (int i = 0; i < numPhysicalDevices; i++) {
            VkPhysicalDeviceProperties currentPhysicalDeviceProperties = {};
            vkGetPhysicalDeviceProperties(physicalDevices[i], &currentPhysicalDeviceProperties);
            if (currentPhysicalDeviceProperties.deviceType == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU && !firstDedicatedPhysicalDeviceIndex.has_value())
                firstDedicatedPhysicalDeviceIndex = i;

            physicalDeviceProperties.push_back(currentPhysicalDeviceProperties);
        }

        // Try dedicated first, then try every GPU in order
        if (firstDedicatedPhysicalDeviceIndex.has_value()) {
            int dedicatedIndex = firstDedicatedPhysicalDeviceIndex.value();
            physicalDevices.insert(physicalDevices.begin(), physicalDevices[dedicatedIndex]);
            physicalDevices.erase(physicalDevices.begin() + dedicatedIndex + 1);
            physicalDeviceProperties.insert(physicalDeviceProperties.begin(), physicalDeviceProperties[dedicatedIndex]);
            physicalDeviceProperties.erase(physicalDeviceProperties.begin() + dedicatedIndex + 1);
        }

        struct PhysicalDeviceQueueFamilyIndices {
            uint32_t graphics, present;
        };

        auto getQueueFamilyIndices = [&](int currentPhysicalDeviceIndex) -> std::optional<PhysicalDeviceQueueFamilyIndices> {
            uint32_t numQueueFamilies = 0;
            vkGetPhysicalDeviceQueueFamilyProperties(physicalDevices[currentPhysicalDeviceIndex], &numQueueFamilies, nullptr);
            std::vector<VkQueueFamilyProperties> queueFamilyProperties(numQueueFamilies);
            vkGetPhysicalDeviceQueueFamilyProperties(physicalDevices[currentPhysicalDeviceIndex], &numQueueFamilies, queueFamilyProperties.data());
            std::optional<uint32_t> graphicsIndex, presentIndex;

            for (int i = 0; i < numQueueFamilies; i++) {
                if (queueFamilyProperties[i].queueFlags & VK_QUEUE_GRAPHICS_BIT && !graphicsIndex.has_value())
                    graphicsIndex = i;

                VkBool32 hasPresentSupport = VK_FALSE;
                vkGetPhysicalDeviceSurfaceSupportKHR(physicalDevices[currentPhysicalDeviceIndex], i, m_surface, &hasPresentSupport);
                if (hasPresentSupport && !presentIndex.has_value())
                    presentIndex = i;
            }

            if (graphicsIndex.has_value() && presentIndex.has_value())
                return PhysicalDeviceQueueFamilyIndices{ graphicsIndex.value(), presentIndex.value() };

            return std::nullopt;
        };

        std::optional<PhysicalDeviceQueueFamilyIndices> chosenPhysicalDeviceIndices;
        for (int i = 0; i < numPhysicalDevices; i++) {
            chosenPhysicalDeviceIndices = getQueueFamilyIndices(i);
            if (chosenPhysicalDeviceIndices.has_value()) {
                // GPU found!
                m_physicalDevice = physicalDevices[i];
                m_queueFIGraphics = chosenPhysicalDeviceIndices->graphics;
                m_queueFIPresent = chosenPhysicalDeviceIndices->present;

                NFLog(std::format("GPU - {}", physicalDeviceProperties[i].deviceName));
                break;
            }
        }

        if (m_physicalDevice == nullptr)
            NFError("No Vulkan GPUs were found to be compatible.");
    }

    void RenderEngine::createDevice() {
        std::set<uint32_t> uniqueQueueFamilyIndices = { m_queueFIGraphics, m_queueFIPresent };

        float queuePriority = 1.0f;
        std::vector<VkDeviceQueueCreateInfo> queueCIs;
        for (uint32_t currentQueueFamilyIndex : uniqueQueueFamilyIndices) {
            VkDeviceQueueCreateInfo queueCI = { VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO };
            queueCI.queueFamilyIndex = currentQueueFamilyIndex;
            queueCI.queueCount = 1;
            queueCI.pQueuePriorities = &queuePriority;
            queueCIs.push_back(queueCI);
        }

        VkPhysicalDeviceFeatures features = {};

        VkDeviceCreateInfo deviceCI = { VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO };
        deviceCI.queueCreateInfoCount = queueCIs.size();
        deviceCI.pQueueCreateInfos = queueCIs.data();
        deviceCI.pEnabledFeatures = &features;

        const char* swapChainExtName = VK_KHR_SWAPCHAIN_EXTENSION_NAME;
        deviceCI.enabledExtensionCount = 1;
        deviceCI.ppEnabledExtensionNames = &swapChainExtName;

        if (vkCreateDevice(m_physicalDevice, &deviceCI, nullptr, &m_device) != VK_SUCCESS)
            NFError("Could not create Vulkan device.");

        vkGetDeviceQueue(m_device, m_queueFIGraphics, 0, &m_queueGraphics);
        vkGetDeviceQueue(m_device, m_queueFIPresent, 0, &m_queuePresent);
    }

    void RenderEngine::createSwapchain() {
        VkSurfaceCapabilitiesKHR surfaceCapabilities = {};
        vkGetPhysicalDeviceSurfaceCapabilitiesKHR(m_physicalDevice, m_surface, &surfaceCapabilities);

        uint32_t numSurfaceFormats = 0;
        vkGetPhysicalDeviceSurfaceFormatsKHR(m_physicalDevice, m_surface, &numSurfaceFormats, nullptr);
        if (numSurfaceFormats == 0)
            NFError("No Vulkan surface formats found.");

        std::vector<VkSurfaceFormatKHR> surfaceFormats(numSurfaceFormats);
        vkGetPhysicalDeviceSurfaceFormatsKHR(m_physicalDevice, m_surface, &numSurfaceFormats, surfaceFormats.data());
        VkSurfaceFormatKHR chosenSurfaceFormat = surfaceFormats[0];
        for (int i = 0; i < surfaceFormats.size(); i++)
            if (surfaceFormats[i].format == VK_FORMAT_B8G8R8A8_SRGB && surfaceFormats[i].colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR)
                chosenSurfaceFormat = surfaceFormats[i];

        uint32_t numPresentModes = 0;
        vkGetPhysicalDeviceSurfacePresentModesKHR(m_physicalDevice, m_surface, &numPresentModes, nullptr);
        if (numPresentModes == 0)
            NFError("No Vulkan surface present modes found.");

        std::vector<VkPresentModeKHR> presentModes(numPresentModes);
        vkGetPhysicalDeviceSurfacePresentModesKHR(m_physicalDevice, m_surface, &numPresentModes, presentModes.data());
        VkPresentModeKHR chosenPresentMode = VK_PRESENT_MODE_FIFO_KHR;
        // TODO: IMMEDIATE isn't always available, so do something about that
        for (int i = 0; i < presentModes.size(); i++)
            if (presentModes[i] == VK_PRESENT_MODE_IMMEDIATE_KHR)
                chosenPresentMode = presentModes[i];

        VkExtent2D swapchainExtent = surfaceCapabilities.currentExtent;
        if (swapchainExtent.width == std::numeric_limits<uint32_t>::max())
            swapchainExtent.width = m_display.width, swapchainExtent.height = m_display.height;

        uint32_t numRequestedImages = surfaceCapabilities.minImageCount + (surfaceCapabilities.minImageCount != surfaceCapabilities.maxImageCount ? 1 : 0);

        VkSwapchainCreateInfoKHR swapchainCI = { VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR };
        swapchainCI.surface = m_surface;
        swapchainCI.minImageCount = numRequestedImages;
        swapchainCI.imageFormat = m_swapchainImageFormat = chosenSurfaceFormat.format;
        swapchainCI.imageColorSpace = chosenSurfaceFormat.colorSpace;
        swapchainCI.imageExtent = swapchainExtent;
        swapchainCI.imageArrayLayers = 1;
        swapchainCI.imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;

        std::vector<uint32_t> queueFamilyIndices = { m_queueFIGraphics, m_queueFIPresent };
        if (m_queueFIGraphics != m_queueFIPresent) {
            swapchainCI.imageSharingMode = VK_SHARING_MODE_CONCURRENT;
            swapchainCI.queueFamilyIndexCount = 2;
            swapchainCI.pQueueFamilyIndices = queueFamilyIndices.data();
        }

        swapchainCI.preTransform = surfaceCapabilities.currentTransform;
        swapchainCI.compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
        swapchainCI.presentMode = chosenPresentMode;
        swapchainCI.clipped = VK_TRUE;

        if (vkCreateSwapchainKHR(m_device, &swapchainCI, nullptr, &m_swapchain) != VK_SUCCESS)
            NFError("Could not create Vulkan swapchain.");

        uint32_t numSwapchainImages = 0;
        vkGetSwapchainImagesKHR(m_device, m_swapchain, &numSwapchainImages, nullptr);
        m_swapchainImages.resize(numSwapchainImages);
        vkGetSwapchainImagesKHR(m_device, m_swapchain, &numSwapchainImages, m_swapchainImages.data());

        m_swapchainImageViews.resize(numSwapchainImages);
        for (int i = 0; i < m_swapchainImageViews.size(); i++) {
            VkImageViewCreateInfo imageViewCI = { VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO };
            imageViewCI.image = m_swapchainImages[i];
            imageViewCI.viewType = VK_IMAGE_VIEW_TYPE_2D;
            imageViewCI.format = m_swapchainImageFormat;
            imageViewCI.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            imageViewCI.subresourceRange.levelCount = 1;
            imageViewCI.subresourceRange.layerCount = 1;

            if (vkCreateImageView(m_device, &imageViewCI, nullptr, &m_swapchainImageViews[i]) != VK_SUCCESS)
                NFError("Could not create Vulkan swapchain image view.");
        }
    }

    void RenderEngine::createOutputRenderPass() {
        VkAttachmentDescription outputColorAttachment = {};
        outputColorAttachment.format = m_swapchainImageFormat;
        outputColorAttachment.samples = VK_SAMPLE_COUNT_1_BIT;
        outputColorAttachment.loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
        outputColorAttachment.storeOp = VK_ATTACHMENT_STORE_OP_STORE;
        outputColorAttachment.stencilLoadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
        outputColorAttachment.stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
        outputColorAttachment.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
        outputColorAttachment.finalLayout = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;

        VkAttachmentReference outputColorAttachmentReference = {};
        outputColorAttachmentReference.attachment = 0;
        outputColorAttachmentReference.layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

        VkSubpassDescription subpassDescription = {};
        subpassDescription.pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
        subpassDescription.colorAttachmentCount = 1;
        subpassDescription.pColorAttachments = &outputColorAttachmentReference;

        VkRenderPassCreateInfo renderPassCI = { VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO };
        renderPassCI.attachmentCount = 1;
        renderPassCI.pAttachments = &outputColorAttachment;
        renderPassCI.subpassCount = 1;
        renderPassCI.pSubpasses = &subpassDescription;

        if (vkCreateRenderPass(m_device, &renderPassCI, nullptr, &m_renderPassOutput) != VK_SUCCESS)
            NFError("Could not create Vulkan output render pass.");
    }

    void RenderEngine::createOutputPipeline() {
        // First, create shader modules
        std::string outputShaderVertexBinary, outputShaderFragmentBinary;
        if (!util::readFile("shaders/output.vert.glsl.spv", outputShaderVertexBinary) || !util::readFile("shaders/output.frag.glsl.spv", outputShaderFragmentBinary))
            NFError("Could not read output shader binaries.");

        ShaderModule outputShaderVertexModule(m_device, outputShaderVertexBinary);
        ShaderModule outputShaderFragmentModule(m_device, outputShaderFragmentBinary);

        // Fill out pipeline shader stages
        VkPipelineShaderStageCreateInfo outputShaderStages[] = { {VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO}, {VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO} };
        outputShaderStages[0].stage = VK_SHADER_STAGE_VERTEX_BIT;
        outputShaderStages[0].module = outputShaderVertexModule.getHandle();
        outputShaderStages[0].pName = outputShaderStages[1].pName = "main";
        outputShaderStages[1].stage = VK_SHADER_STAGE_FRAGMENT_BIT;
        outputShaderStages[1].module = outputShaderFragmentModule.getHandle();

        // Specify dynamic state
        VkDynamicState dynamicState = VK_DYNAMIC_STATE_VIEWPORT;
        VkPipelineDynamicStateCreateInfo dynamicStateCI = { VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO };
        dynamicStateCI.dynamicStateCount = 1;
        dynamicStateCI.pDynamicStates = &dynamicState;

        // Specify vertex input state
        VkPipelineVertexInputStateCreateInfo vertexInputStateCI = { VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO };

        // Specify input assembly state
        VkPipelineInputAssemblyStateCreateInfo inputAssemblyStateCI = { VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO };
        inputAssemblyStateCI.topology = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;

        // Specify viewport state
        VkRect2D scissor = { {0, 0}, {m_display.width, m_display.height} };
        VkPipelineViewportStateCreateInfo viewportStateCI = { VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO };
        viewportStateCI.viewportCount = 1;
        viewportStateCI.scissorCount = 1;
        viewportStateCI.pScissors = &scissor;

        // Specify rasterization state
        VkPipelineRasterizationStateCreateInfo rasterizationCI = { VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO };
        rasterizationCI.lineWidth = 1.0;
        rasterizationCI.cullMode = VK_CULL_MODE_BACK_BIT;
        rasterizationCI.frontFace = VK_FRONT_FACE_CLOCKWISE;

        // Specify multisample state
        VkPipelineMultisampleStateCreateInfo multisampleStateCI = { VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO };
        multisampleStateCI.rasterizationSamples = VK_SAMPLE_COUNT_1_BIT;

        // Specify color blend state
        VkPipelineColorBlendAttachmentState colorBlendAttachmentState = {};
        colorBlendAttachmentState.colorWriteMask = VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT;
        VkPipelineColorBlendStateCreateInfo colorBlendStateCI = { VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO };
        colorBlendStateCI.attachmentCount = 1;
        colorBlendStateCI.pAttachments = &colorBlendAttachmentState;

        // Create pipeline layout
        VkPipelineLayoutCreateInfo layoutCI = { VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO };

        if (vkCreatePipelineLayout(m_device, &layoutCI, nullptr, &m_pipelineLayoutOutput) != VK_SUCCESS)
            NFError("Could not create pipeline layout.");

        // And finally put it all together
        VkGraphicsPipelineCreateInfo pipelineCI = { VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO };
        pipelineCI.stageCount = 2;
        pipelineCI.pStages = outputShaderStages;
        pipelineCI.pVertexInputState = &vertexInputStateCI;
        pipelineCI.pInputAssemblyState = &inputAssemblyStateCI;
        pipelineCI.pViewportState = &viewportStateCI;
        pipelineCI.pRasterizationState = &rasterizationCI;
        pipelineCI.pMultisampleState = &multisampleStateCI;
        pipelineCI.pColorBlendState = &colorBlendStateCI;
        pipelineCI.pDynamicState = &dynamicStateCI;
        pipelineCI.layout = m_pipelineLayoutOutput;
        pipelineCI.renderPass = m_renderPassOutput;

        if (vkCreateGraphicsPipelines(m_device, nullptr, 1, &pipelineCI, nullptr, &m_pipelineOutput) != VK_SUCCESS)
            NFError("Could not create graphics pipeline.");

        // TODO: For now, create swapchain framebuffers
        m_swapchainFramebuffers.resize(m_swapchainImageViews.size());
        for (int i = 0; i < m_swapchainImageViews.size(); i++) {
            VkFramebufferCreateInfo framebufferCI = { VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO };
            framebufferCI.renderPass = m_renderPassOutput;
            framebufferCI.attachmentCount = 1;
            framebufferCI.pAttachments = &m_swapchainImageViews[i];
            framebufferCI.width = m_display.width;
            framebufferCI.height = m_display.height;
            framebufferCI.layers = 1;

            if (vkCreateFramebuffer(m_device, &framebufferCI, nullptr, &m_swapchainFramebuffers[i]) != VK_SUCCESS)
                NFError("Could not create framebuffer.");
        }
    }

    void RenderEngine::createCommandBuffer() {
        VkCommandPoolCreateInfo commandPoolCI = { VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO };
        commandPoolCI.flags = VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
        commandPoolCI.queueFamilyIndex = m_queueFIGraphics;

        if (vkCreateCommandPool(m_device, &commandPoolCI, nullptr, &m_commandPool) != VK_SUCCESS)
            NFError("Could not create command pool.");

        VkCommandBufferAllocateInfo commandBufferAI = { VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO };
        commandBufferAI.commandPool = m_commandPool;
        commandBufferAI.commandBufferCount = 1;

        if (vkAllocateCommandBuffers(m_device, &commandBufferAI, &m_commandBuffer) != VK_SUCCESS)
            NFError("Could not create command buffer.");
    }

    void RenderEngine::createSynchronizationObjects() {
        VkSemaphoreCreateInfo semaphoreCI = { VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO };
        if (vkCreateSemaphore(m_device, &semaphoreCI, nullptr, &m_semaphoreImageAvailable) != VK_SUCCESS ||
            vkCreateSemaphore(m_device, &semaphoreCI, nullptr, &m_semaphoreRenderFinished) != VK_SUCCESS)
            NFError("Could not create semaphore.");

        VkFenceCreateInfo fenceCI = { VK_STRUCTURE_TYPE_FENCE_CREATE_INFO };
        if (vkCreateFence(m_device, &fenceCI, nullptr, &m_fenceInFlight) != VK_SUCCESS)
            NFError("Could not create fence.");
    }

    RenderEngine::~RenderEngine() {
        vkDestroyFence(m_device, m_fenceInFlight, nullptr);
        vkDestroySemaphore(m_device, m_semaphoreRenderFinished, nullptr);
        vkDestroySemaphore(m_device, m_semaphoreImageAvailable, nullptr);

        vkDestroyCommandPool(m_device, m_commandPool, nullptr);

        for (int i = 0; i < m_swapchainFramebuffers.size(); i++)
            vkDestroyFramebuffer(m_device, m_swapchainFramebuffers[i], nullptr);

        vkDestroyPipeline(m_device, m_pipelineOutput, nullptr);
        vkDestroyPipelineLayout(m_device, m_pipelineLayoutOutput, nullptr);
        vkDestroyRenderPass(m_device, m_renderPassOutput, nullptr);

        for (int i = 0; i < m_swapchainImageViews.size(); i++)
            vkDestroyImageView(m_device, m_swapchainImageViews[i], nullptr);

        vkDestroySwapchainKHR(m_device, m_swapchain, nullptr);
        vkDestroyDevice(m_device, nullptr);
        vkDestroySurfaceKHR(m_instance, m_surface, nullptr);
        vkDestroyInstance(m_instance, nullptr);
    }
}
